#include "cffmpegencode.h"

CFFmpegEncode::CFFmpegEncode()
{
    m_opt = NULL;
    m_fmt = NULL;
    m_oc = NULL;
    m_videoCodec = NULL;
    m_audioCodec = NULL;
    m_nHaveAudio = 0;
    m_nHaveVideo = 0;
    memset(&m_videoSt, 0, sizeof(OutputStream));
    memset(&m_audioSt, 0, sizeof(OutputStream));
    memset(&m_options, 0, sizeof(OPTIONS));

    pthread_mutex_init(&m_mutex, NULL);
}

bool CFFmpegEncode::InitFFmpeg(OPTIONS option)
{
    m_options = option;
    int nRet;
    char fileName[256];

    av_register_all();

    if(strlen(m_options.szOutputFileName) == 0)
    {
        printf("You must set '--output' option!\n");
        return false;
    }

    if(m_options.dwMp4 == 1)
    {
        sprintf(fileName, "%s", m_options.szOutputFileName);
    }
    else
        strcpy(fileName, "video0.mp4");

    avformat_alloc_output_context2(&m_oc, NULL, NULL ,fileName);
    if(!m_oc)
        avformat_alloc_output_context2(&m_oc, NULL, "mp4", fileName);

    if(!m_oc)
        return false;

    m_fmt = m_oc->oformat;
    m_fmt->video_codec =AV_CODEC_ID_H264;
    if(m_fmt->video_codec != AV_CODEC_ID_NONE)
    {
        nRet = AddStream(&m_videoSt, m_oc, &m_videoCodec, m_fmt->video_codec);
        if(nRet != 0)
            return false;

        m_nHaveVideo = 1;
    }
    if(m_nHaveVideo)
    {
        OpenVideo(m_videoCodec, &m_videoSt, m_opt);
    }

    m_fmt->audio_codec = AV_CODEC_ID_AAC;
    if(m_fmt->audio_codec != AV_CODEC_ID_NONE)
    {
        nRet = AddStream(&m_audioSt, m_oc, &m_audioCodec, m_fmt->audio_codec);
        if(nRet != 0)
            return false;

        m_nHaveAudio = 1;
    }
    if(m_nHaveAudio)
    {
        bool bRet = OpenAudio(m_oc, m_audioCodec, &m_audioSt, m_opt);
        if(bRet != true)
            return false;
    }

    av_dump_format(m_oc, 0, fileName, 1);

    //open output file if needed
    if(!(m_fmt->flags & AVFMT_NOFILE))
    {
        nRet = avio_open(&m_oc->pb, fileName, AVIO_FLAG_WRITE);
        if(nRet < 0)
            return false;
    }

    //write stream header
    nRet = avformat_write_header(m_oc, &m_opt);
    if(nRet < 0)
        return false;

    return true;
}

void CFFmpegEncode::CloseFFmpeg()
{
    av_write_trailer(m_oc);

    if(m_nHaveVideo)
    {
        CloseStream(m_oc, &m_videoSt);
    }

    if(m_nHaveAudio)
    {
        CloseStream(m_oc, &m_audioSt);
    }

    if(!(m_fmt->flags & AVFMT_NOFILE))
    {
        avio_closep(&m_oc->pb);
    }

    avformat_free_context(m_oc);
    pthread_mutex_destroy(&m_mutex);
}

bool CFFmpegEncode::VideoEncodeCallback(uint8_t *pImage, bool bAudioCapture)
{
    if(pImage == NULL)
        return false;

    int rc = 0;
    if(!bAudioCapture || av_compare_ts(m_videoSt.next_pts, m_videoSt.st->codec->time_base,
                        m_audioSt.next_pts, m_audioSt.st->codec->time_base) <= 0)
    {
        pthread_mutex_lock(&m_mutex);
        rc = WriteVideoFrame(m_oc, &m_videoSt, pImage);
        pthread_mutex_unlock(&m_mutex);

        if(rc != 0)
            return false;
    }

    return true;
}

bool CFFmpegEncode::AudioEncodeCallback(uint8_t *pAudioBuf)
{
    if(pAudioBuf == NULL)
        return false;

    int rc = 0;
    pthread_mutex_lock(&m_mutex);
    rc = WriteAudioFrame(m_oc, &m_audioSt, pAudioBuf);
    pthread_mutex_unlock(&m_mutex);

    if(rc != 0)
        return false;

    return true;
}

void CFFmpegEncode::FlushAudioEncodeCallback()
{
    AVCodecContext *c;
    c = m_audioSt.st->codec;
    int ret;
    int got_frame;
    AVPacket enc_pkt = {0};
    while (1) {
        av_init_packet(&enc_pkt);
        ret = avcodec_encode_audio2 (c, &enc_pkt, NULL, &got_frame);
        av_frame_free(NULL);

        if(!got_frame)
            break;

        ret = WriteFrame(m_oc, &c->time_base, m_audioSt.st, &enc_pkt);
        if(ret < 0)
            break;
    }
}

void CFFmpegEncode::CloseStream(AVFormatContext *oc, OutputStream *ost)
{
    avcodec_close(ost->st->codec);
    av_frame_free(&ost->frame);
    av_frame_free(&ost->tmp_frame);
    swr_free(&ost->swr_ctx);
}

//write frame
int CFFmpegEncode::WriteFrame(AVFormatContext *fmt_ctx, const AVRational *time_base, AVStream *st, AVPacket *pkt)
{
    av_packet_rescale_ts(pkt, *time_base, st->time_base);
    pkt->stream_index = st->index;

    return av_interleaved_write_frame(fmt_ctx, pkt);
}

//add stream
int CFFmpegEncode::AddStream(OutputStream *ost, AVFormatContext *oc,
                       AVCodec **codec, enum AVCodecID codec_id)
{
    AVCodecContext *c;

    //fine the encoder
    *codec = avcodec_find_encoder(codec_id);
    if(!(*codec))
        return 1;

    ost->st = avformat_new_stream(oc, *codec);
    if(!ost->st)
        return 1;

    ost->st->id = oc->nb_streams - 1;
    c = ost->st->codec;

    switch((*codec)->type)
    {
    case AVMEDIA_TYPE_AUDIO:
    {
        c->sample_fmt = (*codec)->sample_fmts ? (*codec)->sample_fmts[0] : AV_SAMPLE_FMT_S32P;
        c->bit_rate = 64000;
        c->sample_rate = 44100;
        c->channel_layout = AV_CH_LAYOUT_STEREO;
        c->channels = av_get_channel_layout_nb_channels(c->channel_layout);
        ost->st->time_base = (AVRational){1, c->sample_rate};
        break;
    }

    case AVMEDIA_TYPE_VIDEO:
    {
        c->codec_id = codec_id;

        if(m_options.dwBitrate >= 0)
        {
            c->bit_rate = m_options.dwBitrate * 1000;
        }
        else
        {
            c->bit_rate = 4000000;
        }

        c->width = m_options.dwWidth;
        c->height = m_options.dwHeight;

        av_opt_set(c->priv_data, "preset", "ultrafast", 0);
        av_opt_set(c->priv_data, "tune", "zerolatency", 0);

        switch (m_options.dwProfile)
        {
        case 0:
            c->profile = FF_PROFILE_H264_HIGH;
            break;
        case 1:
            c->profile = FF_PROFILE_H264_MAIN;
            break;
        case 2:
            c->profile = FF_PROFILE_H264_BASELINE;
            break;
        default:
            break;
        }

        int nFrameRate = 0;
        nFrameRate = m_options.dwFPS;
        ost->st->time_base = (AVRational){1, nFrameRate};
        c->time_base = ost->st->time_base;

        if(m_options.dwBFrame > 0)
        {
            c->max_b_frames = m_options.dwBFrame;
        }

        if(m_options.dwIFrames > 0)
        {
            c->gop_size = m_options.dwIFrames;
        }
        else
            c->gop_size = 12;

        c->pix_fmt = STREAM_PIX_FMT;
        break;
    }

    default:
        break;
    }

    if(oc->oformat->flags & AVFMT_GLOBALHEADER)
    {
        c->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
    }

    return 0;
}

//video output
AVFrame *CFFmpegEncode::AllocPicture(enum AVPixelFormat pix_fmt, int width, int height)
{
    AVFrame *picture;
    int ret;

    picture = av_frame_alloc();
    if(picture == NULL)
        return NULL;

    picture->format = pix_fmt;
    picture->width  = width;
    picture->height = height;

    /* allocate the buffers for the frame data */
    ret = av_frame_get_buffer(picture, 1);
    if (ret < 0)
        return NULL;

    return picture;
}

bool CFFmpegEncode::OpenVideo(AVCodec *codec, OutputStream *ost, AVDictionary *opt_arg)
{
    int ret;
    AVCodecContext *c = ost->st->codec;
    AVDictionary *opt = NULL;
    av_dict_copy(&opt, opt_arg, 0);

    //open the codec
    ret = avcodec_open2(c, codec, &opt);
    av_dict_free(&opt);
    if(ret < 0)
        return false;

    //allocate and init a re_usable frame
    ost->frame = AllocPicture(c->pix_fmt, c->width, c->height);
    if(!ost->frame)
        return false;

    return true;
}

AVFrame *CFFmpegEncode::GetVideoFrame(OutputStream *ost, uint8_t *imageData)
{
    if(imageData == NULL)
        return NULL;


    ost->frame->data[0] = imageData;
    ost->frame->data[1] = imageData + m_options.dwWidth * m_options.dwHeight;
    ost->frame->data[2] = imageData + m_options.dwWidth * m_options.dwHeight * 5 / 4;

    ost->frame->pts = ost->next_pts++;

    return ost->frame;
}

int CFFmpegEncode::WriteVideoFrame(AVFormatContext *oc, OutputStream *ost, uint8_t *imageData)
{
    int ret;
    AVCodecContext *c;
    AVFrame *frame;
    int got_packet = 0;
    AVPacket pkt = {0};
    c = ost->st->codec;
    frame = GetVideoFrame(ost, imageData);
    av_init_packet(&pkt);

    //encode the image
    ret = avcodec_encode_video2(c, &pkt, frame, &got_packet);
    if(ret < 0)
        return 1;

    if(got_packet)
        ret = WriteFrame(oc, &c->time_base, ost->st, &pkt);
    else
        ret = 0;

    if(ret < 0)
        return 1;

    return 0;
}

//audio
AVFrame *CFFmpegEncode::AllocAudioFrame(enum AVSampleFormat sample_fmt, uint64_t channel_layout,
                                  int sample_rate, int nb_samples)
{
    AVFrame *frame = av_frame_alloc();
    int ret;

    if(!frame)
        return NULL;

    frame->format = sample_fmt;
    frame->channel_layout = channel_layout;
    frame->sample_rate = sample_rate;
    frame->nb_samples = nb_samples;

    if(nb_samples)
    {
        ret = av_frame_get_buffer(frame, 0);
        if(ret < 0)
            return NULL;
    }

    return frame;
}

bool CFFmpegEncode::OpenAudio(AVFormatContext *oc, AVCodec *codec, OutputStream *ost, AVDictionary *opt_arg)
{
    AVCodecContext *c;
    int nb_samples;
    int ret;
    AVDictionary *opt = NULL;
    c = ost->st->codec;

    //open it
    av_dict_copy(&opt, opt_arg, 0);
    ret = avcodec_open2(c, codec, &opt);
    av_dict_free(&opt);
    if(ret < 0)
        return false;

    if(c->codec->capabilities & AV_CODEC_CAP_VARIABLE_FRAME_SIZE)
        nb_samples = 10000;
    else
         nb_samples = c->frame_size;

    ost->frame = AllocAudioFrame(c->sample_fmt, c->channel_layout, c->sample_rate, nb_samples);
    ost->tmp_frame = AllocAudioFrame(AV_SAMPLE_FMT_S32, c->channel_layout, c->sample_rate, nb_samples);
    if(ost->frame == NULL || ost->tmp_frame == NULL)
        return false;

    //create resampler context
    ost->swr_ctx = swr_alloc();
    if(!ost->swr_ctx)
        return false;

    //set options
    av_opt_set_int          (ost->swr_ctx, "in_channel_count", c->channels,         0);
    av_opt_set_int          (ost->swr_ctx, "in_sample_rate",   c->sample_rate,      0);
    av_opt_set_sample_fmt   (ost->swr_ctx, "in_sample_fmt",    AV_SAMPLE_FMT_S32,   0);
    av_opt_set_int          (ost->swr_ctx, "out_channel_count", c->channels,        0);
    av_opt_set_int          (ost->swr_ctx, "out_sample_rate",   c->sample_rate,     0);
    av_opt_set_sample_fmt   (ost->swr_ctx, "out_sample_fmt",    c->sample_fmt,      0);

    if((ret = swr_init(ost->swr_ctx)) < 0)
        return false;

    return true;
}

AVFrame *CFFmpegEncode::GetAudioFrame(OutputStream *ost, uint8_t *pAudioBuf)
{
    if(pAudioBuf == NULL)
        return NULL;

    AVFrame *frame = ost->tmp_frame;
    frame->data[0] = pAudioBuf;
    frame->pts = ost->next_pts;
    ost->next_pts += frame->nb_samples;

    return frame;
}

int CFFmpegEncode::WriteAudioFrame(AVFormatContext *oc, OutputStream *ost, uint8_t *audio_buf)
{
    AVCodecContext *c;
    AVPacket pkt = { 0 };
    AVFrame *frame;
    int ret;
    int got_packet;
    int dst_nb_samples;

    av_init_packet(&pkt);
    c = ost->st->codec;

    frame = GetAudioFrame(ost, audio_buf);
    if(frame)
    {
        dst_nb_samples = av_rescale_rnd(swr_get_delay(ost->swr_ctx, c->sample_rate) + frame->nb_samples,
                                        c->sample_rate, c->sample_rate, AV_ROUND_DOWN);
        av_assert0(dst_nb_samples == frame->nb_samples);
        ret = av_frame_make_writable(ost->frame);
        if(ret < 0)
            return 1;

        ret = swr_convert(ost->swr_ctx, ost->frame->data, dst_nb_samples, (const uint8_t **)frame->data, frame->nb_samples);
        if(ret < 0)
            return 1;

        frame = ost->frame;
        frame->pts = av_rescale_q(ost->samples_count, (AVRational){1, c->sample_rate}, c->time_base);
        ost->samples_count += dst_nb_samples;
    }

    ret = avcodec_encode_audio2(c, &pkt, frame, &got_packet);
    if(ret < 0)
        return 1;

    if(got_packet)
    {
        ret = WriteFrame(oc, &c->time_base, ost->st, &pkt);
        if(ret < 0)
            return 1;
    }

    return 0;
}
