////////////////////////////////////////////////////////////////////////////////
// CONFIDENTIAL and PROPRIETARY software of Magewell Electronics Co., Ltd.
// Copyright (c) 2011-2014 Magewell Electronics Co., Ltd. (Nanjing)
// All rights reserved.
// This copyright notice MUST be reproduced on all authorized copies.
////////////////////////////////////////////////////////////////////////////////
#include "videorenderbase.h"
#include "videorenderxcb.h"
#include "videorendernull.h"
#include "videorenderfb.h"

#include "v4l2capturedevice.h"

#include <string>

#include <stdlib.h>
#include <getopt.h>

int debug_level = 8;

static int g_width = 1280;
static int g_height = 720;
static char *g_dev_name = (char *)"/dev/video0";
static int g_memory_type = 1;
static int g_numerator = 10000000;
static int g_denominator = 333333;
static char *g_format_str = (char *)"YUY2";
#ifndef DISABLE_XCB
static char *g_vo_str = (char *)"xv";
#else
static char *g_vo_str = (char *)"fb";
#endif
static char *g_fb_dev = (char *)"/dev/fb0";
static int g_tty_nr = -1;

struct MWFourccNameMap {
    char        *name;
    uint32_t    fourcc;
};

struct MWFourccNameMap g_fourcc_name_map[] = {
    { (char *)"RGB24",      MWFOURCC_RGB24},
    { (char *)"RGBA",       MWFOURCC_RGBA},
    { (char *)"BGR24",      MWFOURCC_BGR24},
    { (char *)"BGRA",       MWFOURCC_BGRA},
    { (char *)"ARGB",       MWFOURCC_ARGB},
    { (char *)"YUY2",       MWFOURCC_YUY2},
    { (char *)"UYVY",       MWFOURCC_UYVY},
    { (char *)"I420",       MWFOURCC_I420},
    { (char *)"YV12",       MWFOURCC_YV12},
    { (char *)"NV12",       MWFOURCC_NV12},

    { NULL, 0}
};

static char *mwfourcc_to_string(uint32_t fourcc)
{
    for (int i = 0; ; i++) {
        if (g_fourcc_name_map[i].name == NULL) {
            return NULL;
        }

        if (g_fourcc_name_map[i].fourcc == fourcc)
            return g_fourcc_name_map[i].name;
    }

    return NULL;
}

static uint32_t mwstring_to_fourcc(char *name)
{
    for (int i = 0; ; i++) {
        if (g_fourcc_name_map[i].name == NULL)
            return 0;

        if (std::string(g_fourcc_name_map[i].name) ==
                std::string(name)) {
            return g_fourcc_name_map[i].fourcc;
        }
    }

    return 0;
}

static void parse_arg(int argc, char **argv);

static void exit_clean(void *data)
{
    V4l2CaptureDevice *v4l2device = reinterpret_cast<V4l2CaptureDevice*>(data);

    mw_debug(0, "\n");

    v4l2device->StopStreaming();
    v4l2device->DestroyBuffers();
}

int main(int argc, char *argv[])
{
    VideoRenderBase *videoRender = NULL;
    V4l2CaptureDevice *v4l2device = NULL;
    MWFrameFormat format;
    uint32_t fourcc;

    parse_arg(argc, argv);

    if (strcmp(g_vo_str, "null") == 0) {
        videoRender = new VideoRenderNull;
#ifndef DISABLE_XCB
    } else if (strcmp(g_vo_str, "xv") == 0) {
        VideoRenderXCB *_render = new VideoRenderXCB;
        if (!_render->Init(g_width, g_height)) {
            fprintf(stderr, "Video render xv init failed\n");
            exit(-1);
        }
        videoRender = _render;
#endif
    } else if (strcmp(g_vo_str, "fb") == 0) {
        VideoRenderFB *_render = new VideoRenderFB;
        if (!_render->Init(g_fb_dev, g_tty_nr)) {
            fprintf(stderr, "Video render fb(%s) init failed\n", g_fb_dev);
            exit(-1);
        }
        if (!_render->GetGeometry(g_width, g_height)) {
            fprintf(stderr, "Get fb geometry failed\n");
            exit(-1);
        }
        videoRender = _render;
    } else {
        fprintf(stderr, "Unsupported video render: %s\n", g_vo_str);
        exit(-1);
    }

    if (videoRender == NULL)
        return -1;

    v4l2device = new V4l2CaptureDevice(videoRender);

    if (!v4l2device->Init(g_dev_name)) {
        fprintf(stderr, "V4l2 device %s init failed\n", g_dev_name);
        exit(-1);
    }

    if (g_format_str != NULL) {
        if (strcmp(g_format_str, "v4l2") == 0) {
            uint32_t *fourccArray;
            int fourccCount;
            fourccArray = v4l2device->GetSupportedFourcc(fourccCount);

            printf("Available formats[v4l2]:\n");
            for (int i = 0; i < fourccCount; i++) {
                char *fourccName = mwfourcc_to_string(fourccArray[i]);
                if (fourccName != NULL)
                    printf("\t%s\n", fourccName);
            }
            exit(0);
        }

        if (strcmp(g_format_str, "vo") == 0) {
            uint32_t *fourccArray;
            int fourccCount;
            fourccArray = videoRender->GetSupportedFourcc(fourccCount);
            if (fourccArray == NULL &&
                    std::string(videoRender->GetName()) == "null") {
                printf("Video render support all formats!\n");
                exit(0);
            }

            if (fourccCount <= 0) {
                printf("No available format for %s\n", videoRender->GetName());
            } else {
                printf("Available formats[%s]:\n", videoRender->GetName());
                for (int i = 0; i < fourccCount; i++) {
                    char *fourccName = mwfourcc_to_string(fourccArray[i]);
                    if (fourccName != NULL)
                        printf("\t%s\n", fourccName);
                }
            }
            exit(0);
        }
    }

    fourcc = mwstring_to_fourcc(g_format_str);
    if (fourcc == 0) {
        fprintf(stderr, "Unsupported format: %s\n", g_format_str);
        exit(-1);
    }

    if (!videoRender->SetFrameFormat(g_width, g_height, fourcc)) {
        fprintf(stderr, "[vo] Unsupported format: %dx%d(%s)\n",
                g_width, g_height, g_format_str);
        exit(-1);
    }

    format.width = g_width;
    format.height = g_height;
    format.fourcc = fourcc;
    format.denominator = g_numerator;
    format.numerator = g_denominator;
    if (!v4l2device->SetFrameFormat(format)) {
        fprintf(stderr, "[v4l2] Unsupported format: %dx%d(%s)\n",
                g_width, g_height, g_format_str);
        exit(-1);
    }

    if (v4l2device->CreateBuffers(
                4, g_memory_type == 0 ? V4L2_MEMORY_MMAP : V4L2_MEMORY_USERPTR) <= 0) {
        fprintf(stderr, "[v4l2] create buffers failed\n");
        exit(-1);
    }

    printf("Video capture device: %s\n", g_dev_name);
    printf("V4l2 memory type: %s\n", g_memory_type == 0 ? "mmap" : "userptr");
    printf("Video width: %d\n", g_width);
    printf("Video height: %d\n", g_height);
    printf("Video format: %s\n", mwfourcc_to_string(fourcc));
    if (strcmp(g_vo_str, "fb") == 0) {
        printf("Video output: %s(%s)\n", g_vo_str, g_fb_dev);
    } else {
        printf("Video output: %s\n", g_vo_str);
    }
    printf("\n");

    if (!v4l2device->StartStreaming()) {
        fprintf(stderr, "[v4l2] start streaming failed\n");
        exit(-1);
    }

    videoRender->RunLoop(exit_clean, v4l2device);

    return 0;
}

static void usage(FILE *fp, int argc, char **argv)
{
    fprintf(fp,
            "USAGE\n"
            "  mwcap-v4l2 [OPTIONS]\n"
            "\n"
            "DESCRIPTION\n"
            "  Demonstrate the video4linux2 video capture.\n"
            "\n"
            "OPTIONS\n"
            "  -h, --help                       print this message\n"
            "  -d, --device <device>            select v4l2 device(default:/dev/video0)\n"
            "  --width <width>                  set video width\n"
            "  --height <height>                set video height\n"
            "  -f, --format <format>            set video format(default:YUY2)\n"
            "                                   <v4l2> list available formats for v4l2 device\n"
            "                                   <vo>list of available formats for video\n"
            "                                     output device\n"
            "  -m, --memory <0|1>               select v4l2 memory type (0 : mmap, 1 : usrptr(default))\n"
            "  -r, --framerate <numerator/denominator>\n"
            "                                   set video frame rate (eg. 30/1 for 30 fps)\n"
            "  -o, --vo <render>                select video output render(xv(default), fb or null)\n"
            "  -b, --framebuffer <device>       select framebuffer device(default:/dev/fb0)\n"
            "\n");
}

static const char short_options[] = "hd:m:f:r:o:b:";

static const struct option long_options[] = {
    { "help",               no_argument,       NULL, 'h' },
    { "device",             required_argument, NULL, 'd' },
    { "memory",             required_argument, NULL, 'm' },
    { "width",              required_argument, NULL, 'w' },
    { "height",             required_argument, NULL, 't' },
    { "format",             required_argument, NULL, 'f' },
    { "framerate",          required_argument, NULL, 'r' },
    { "vo",                 required_argument, NULL, 'o' },
    { "framebuffer",        required_argument, NULL, 'b' },

    { 0, 0, 0, 0 }
};

static void parse_arg(int argc, char **argv)
{
    for (;;) {
        int idx;
        int c;

        c = getopt_long(argc, argv,
                        short_options, long_options, &idx);

        if (-1 == c)
            break;

        switch (c) {
        case 0: /* getopt_long() flag */
            break;

        case 'h':
            usage(stdout, argc, argv);
            exit(0);

        case 'd':
            g_dev_name = optarg;
            if (g_dev_name == NULL) {
                usage(stdout, argc, argv);
                exit(-1);
            }
            break;

        case 'm':
            g_memory_type = strtol(optarg, NULL, 0);
            if (g_memory_type != 0 && g_memory_type != 1) {
                usage(stdout, argc, argv);
                exit(-1);
            }
            break;

        case 'w':
            g_width = strtol(optarg, NULL, 0);
            break;
        case 't':
            g_height = strtol(optarg, NULL, 0);
            break;

        case 'r':
            sscanf(optarg, "%d/%d", &g_numerator, &g_denominator);
            break;

        case 'f':
            g_format_str = optarg;
            if (g_format_str == NULL) {
                usage(stdout, argc, argv);
                exit(-1);
            }
            break;

        case 'o':
            g_vo_str = optarg;
            if (g_vo_str == NULL) {
                usage(stdout, argc, argv);
                exit(-1);
            }
            break;

        case 'b':
            g_fb_dev = optarg;
            if (g_fb_dev == NULL) {
                usage(stdout, argc, argv);
                exit(-1);
            }
            break;

        default:
            usage(stdout, argc, argv);
            exit(0);
        }
    }

    if (g_width == 0 || g_height == 0) {
        usage(stdout, argc, argv);
        exit(0);
    }
}
